/*
 * KernelConf - A Linux Kernel Configuration System
 *
 * Distributed under GNU General Public Lisence
 *
 * Copyright (C) 2002 by Anuradha Ratnaweera
 */

#include <ctype.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>

#include "expressions.h"
#include "symbols.h"
#include "utils.h"
#include "values.h"

extern SymbolPtr symbol_tree_root;

/*
 * Read the next token from the expression and returns its type
 * precedence and length.  Apart from the operator types, SYMBOL
 * and VALUE types are recognized.  Symbols contain alphabatical,
 * numerical and underscore characters and values are contained
 * within quotes.  TYPE_NULL is returned if we are at the end of
 * the string.
 */
static char *get_next_token(char* expr, TokenPtr token, size_t *length)
{
    char *p;

    while (isspace(*expr))
	expr++;

    switch (*expr) {
	case 0:
	    *length = 1;
	    token->type = TYPE_NULL;
	    token->precedence = PRE_NULL;
	    return expr;
	case '<':
	    switch (*(expr + 1)) {
		case '-':
		    *length = 2;
		    token->type = TYPE_DERIVED;
		    token->precedence = PRE_ASSIGN;
		    return expr;
		case '=':
		    *length = 2;
		    token->type = TYPE_LTE;
		    token->precedence = PRE_REL;
		    return expr;
		default:
		    *length = 1;
		    token->type = TYPE_LT;
		    token->precedence = PRE_REL;
		    return expr;
	    }
	case '>':
	    switch (*(expr + 1)) {
		case '=':
		    *length = 2;
		    token->type = TYPE_GTE;
		    token->precedence = PRE_REL;
		    return expr;
		default:
		    *length = 1;
		    token->type = TYPE_GT;
		    token->precedence = PRE_REL;
		    return expr;
	    }
	case '(':
	    *length = 1;
	    token->type = TYPE_LEFT_PARA;
	    token->precedence = PRE_NULL;
	    return expr;
	case ')':
	    *length = 1;
	    token->type = TYPE_RIGHT_PARA;
	    token->precedence = PRE_NULL;
	    return expr;
	case '!':
	    switch (*(expr + 1)) {
		case '=':
		    *length = 2;
		    token->type = TYPE_NOT_EQUAL;
		    token->precedence = PRE_EQUAL;
		    return expr;
		default:
		    *length = 1;
		    token->type = TYPE_NOT;
		    token->precedence = PRE_NOT;
		    return expr;
	    }
	case '=':
	    switch (*(expr + 1)) {
		case '=':
		    *length = 2;
		    token->type = TYPE_EQUAL;
		    token->precedence = PRE_EQUAL;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }
	case '&':
	    switch (*(expr + 1)) {
		case '&':
		    *length = 2;
		    token->type = TYPE_AND;
		    token->precedence = PRE_AND;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }
	case '|':
	    switch (*(expr + 1)) {
		case '|':
		    *length = 2;
		    token->type = TYPE_OR;
		    token->precedence = PRE_OR;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }

	case '-':
	    switch (*(expr + 1)) {
		case '>':
		    *length = 2;
		    token->type = TYPE_DEPENDS;
		    token->precedence = PRE_ASSIGN;
		    return expr;
		default:
		    *length = 0;
		    token->type = TYPE_UNKNOWN;
		    token->precedence = PRE_NULL;
		    return expr;
	    }
	case '"':
	    p = ++expr;
	    while ((*p) && ((*p) != '"'))
		p++;
	    if (*p) {
		*p = ' ';
		*length = p - expr;
		token->type = TYPE_VALUE;
		token->precedence = PRE_NULL;
		return expr;
	    }
	    else {
		*length = 0;
		token->type = TYPE_UNKNOWN;
		token->precedence = PRE_NULL;
		return expr;
	    }
	default:
	    p = expr;
	    while (isalnum(*p) || ((*p) == '_'))
		p++;
	    if (p == expr) {
		*length = 0;
		token->type = TYPE_UNKNOWN;
		token->precedence = PRE_NULL;
		return expr;
	    }
	    else {
		*length = p - expr;
		token->type = TYPE_SYMBOL;
		token->precedence = PRE_NULL;
		return expr;
	    }
    }

    *length = 0;
    token->type = TYPE_UNKNOWN;
    token->precedence = PRE_NULL;
    return expr;
}

/*
 * Pushes the TOKEN into the stack.  *STACK == NULL if empty.
 */
static void push(TokenPtr *stack, TokenPtr token)
{
    token->next = *stack;
    *stack = token;
}

/*
 * Pops from the stack and returns a pointer to the top element.
 * Returns NULL if the stack is empty.
 */
static TokenPtr pop(TokenPtr *stack)
{
    TokenPtr token;

    if (*stack) {
	token = *stack;
	*stack = (*stack)->next;
	token->next = NULL;
	return token;
    }
    else
	return NULL;
}

#ifdef DEBUG0

static void show_list(TokenPtr list, char *str)
{
    TokenPtr token;
    printf("%s: ", str);
    for (token = list; token; token = token->next)
	if (token->type == TYPE_VALUE)
	    printf("\"%s\" ", (char*)(token->value.strval));
	else if (token->type == TYPE_SYMBOL)
	    printf("*%s* ", ((NodePtr)(token->value.ptr))->name);
	else if (token->type == TYPE_BOOL)
	    printf("*%c* ", token->value.charval);
	else if (token->type == TYPE_TRI)
	    printf("*%c* ", token->value.charval);
	else
	    printf("%d ", token->type);
    printf("\n");
}

#endif

/*
 * Adds TOKEN to the end of a linked list pointed by FIRST whose last
 * element is LAST.
 */
static void add_token_to_list(TokenPtr *first, TokenPtr *last, TokenPtr token)
{
    if (*first)
	(*last)->next = token;
    else
	*first = token;

    *last = token;
	token->next = NULL;
}

/*
 * Releases memory used by token and if necessary by its elements.
 */
static void free_token(TokenPtr token)
{
    if ((token->type == TYPE_VALUE) || (token->type == TYPE_STR))
	free(token->value.strval);
    free(token);
}

/*
 * Parses the string EXPR related to SYMBOL and returns an expression
 * which contains a list of tokens in reverse Polish.
 *
 * If DEPENDENCY is nonzero, it does the followings, too:
 *
 * - If EXPR doesn't contain a dependency operator EXPR is replaced
 *   with S <= EXPR.
 * - If EXPR is of the form a -> A, it is replaced with (s == a) -> A.
 *
 *   Here a is a value, A is any sub-expression and s is the value
 *   of the current symbol.
 */
static ExpressionPtr parse_expression(char *expr, NodePtr symbol, int dependency)
{
    ExpressionPtr expression;
    TokenPtr stack, token, last_token, tmp;
    size_t length;
    char c, have_dependency = 0;

    expression = malloc(sizeof(Expression));
    expression->first_token = NULL;
    expression->next = NULL;

    stack = NULL;

    do {
	token = (TokenPtr)malloc(sizeof(Token));
	expr = get_next_token(expr, token, &length);

	switch (token->type) {
	    case TYPE_LEFT_PARA:
		push(&stack, token);
		break;
	    case TYPE_RIGHT_PARA:
		while (stack && (stack->type != TYPE_LEFT_PARA)) {
		    tmp = pop(&stack);
		    add_token_to_list(&(expression->first_token),
			    &last_token, tmp);
#ifdef DEBUG0
		    show_list(stack, "Stack");
		    show_list(expression->first_token, "Expression");
#endif
		}
		pop(&stack);
		break;
	    case TYPE_DEPENDS:
		have_dependency = 1;
		/* don't break */
	    case TYPE_OR:
	    case TYPE_AND:
	    case TYPE_EQUAL:
	    case TYPE_NOT_EQUAL:
	    case TYPE_LT:
	    case TYPE_GT:
	    case TYPE_LTE:
	    case TYPE_GTE:
	    case TYPE_NOT:
	    case TYPE_DERIVED:
		while (stack && (stack->precedence >= token->precedence)) {
		    tmp = pop(&stack);
		    add_token_to_list(&(expression->first_token),
			    &last_token, tmp);
#ifdef DEBUG0
		    show_list(stack, "Stack");
		    show_list(expression->first_token, "Expression");
#endif
		}
		push(&stack, token);
		break;
	    case TYPE_SYMBOL:
		c = expr[length];
		expr[length] = 0;
		token->value.ptr = (void*)search_tree((NodePtr*)(&symbol_tree_root),
			expr, NODE_MARK_INSERT, sizeof(Symbol), NULL);
		add_token_to_list(&(expression->first_token),
			&last_token, token);
		expr[length] = c;
		break;
	    case TYPE_VALUE:
		c = expr[length];
		expr[length] = 0;
		token->value.strval = string_duplicate(expr);
		add_token_to_list(&(expression->first_token),
			&last_token, token);
		expr[length] = c;
		break;
	    case TYPE_UNKNOWN:
		return NULL;
	    case TYPE_NULL:
		/* why doesn't the following line work?
		free(token); */
		break;
	    default:
		;
	}

#ifdef DEBUG0
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif

	expr += length;
    } while (token->type != TYPE_NULL);

    while (stack) {
	tmp = pop(&stack);
	add_token_to_list(&(expression->first_token),
		&last_token, tmp);
#ifdef DEBUG0
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif
    }

    /* if expression A doesn't have ->, convert to form s <= A */

    if (dependency && (have_dependency == 0)) {
	/* add current symbol to the beginning */
	token = (TokenPtr)malloc(sizeof(Token));
	token->precedence = PRE_NULL;
	token->type = TYPE_SYMBOL;
	token->value.ptr = (void*)symbol;
	token->next = expression->first_token;
	expression->first_token = token;
#ifdef DEBUG0
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif

	/* add <= to the end */
	token = (TokenPtr)malloc(sizeof(Token));
	token->precedence = PRE_REL;
	token->type = TYPE_LTE;
	token->value.ptr = NULL;
	add_token_to_list(&(expression->first_token),
		&last_token, token);
#ifdef DEBUG0
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif
    }

    /* if expression is of the form a -> A, convert it to (s == a) -> A */

    if (dependency && (expression->first_token->type == TYPE_VALUE) &&
	    (last_token->type == TYPE_DEPENDS)) {
	/* add == to after the first token */
	token = (TokenPtr)malloc(sizeof(Token));
	token->precedence = PRE_EQUAL;
	token->type = TYPE_EQUAL;
	token->next = expression->first_token->next;
	expression->first_token->next = token;
#ifdef DEBUG0
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif

	/* add s to the beginning */
	token = (TokenPtr)malloc(sizeof(Token));
	token->precedence = PRE_NULL;
	token->type = TYPE_SYMBOL;
	token->value.ptr = (void*)symbol;
	token->next = expression->first_token;
	expression->first_token = token;
#ifdef DEBUG0
	show_list(stack, "Stack");
	show_list(expression->first_token, "Expression");
#endif
    }

    return expression;
}

/*
 * Adds EXPRESSION to a linked list whose first element is FIRST.
 * Last pointer's next element is NOT set to NULL.
 */
void add_expressions_to_list(ExpressionPtr *first, ExpressionPtr expressions)
{
    if (expressions == NULL)
	return;

    if (*first) {
	while ((*first)->next)
	    *first = (*first)->next;
	(*first)->next = expressions;
    }
    else
	*first = expressions;
}

/*
 * Accepts a semicolon seperated list of expressions EXPR and parses
 * each one by repeatedly calling parse_expression.  Returns a linked
 * list of expressions, each containing a reverse polish representation
 * of the corresponding expressions.
 */
ExpressionPtr parse_expressions(char *expr, NodePtr symbol, int dependency)
{
    char *c;
    ExpressionPtr expression_list = NULL, last_expression, expression;

    do {
	c = strchr(expr, ';');
	if (c)
	    *c = 0;
#ifdef DEBUG0
	printf("Evaluating: *%s*\n", expr);
#endif
	expression = parse_expression(expr, symbol, dependency);
	if (expression) {
	    if (expression_list) {
		last_expression->next = expression;
		last_expression = last_expression->next;
	    }
	    else {
		expression_list = expression;
		last_expression = expression;
		expression_list->next = NULL;
	    }
	}

	if (c)
	    expr = c + 1;
    } while (c);
    
    return expression_list;
}

/****************************************************************/
/***                end of parsing expressions                ***/
/***              now we've got to evaluate them              ***/
/****************************************************************/

/*
 * This stuff is a bit quick and dirty.  However, they don't
 * effect the other parts of the program so can be, or rather
 * must be ;), rewritten.
 */
static int is_boolean(TokenPtr token)
{
    if (((token->type == TYPE_BOOL) ||
		(token->type == TYPE_TRI) ||
		(token->type == TYPE_OPTION) ||
		(token->type == TYPE_MODULE)) &&
	    ((token->value.charval == 'n') ||
	     (token->value.charval == 'm') ||
	     (token->value.charval == 'y')))
	return 1;
    return 0;
}

static int is_quasi_boolean(TokenPtr token)
{
    if ((token->type == TYPE_VALUE) &&
	    ((token->value.strval[0] == 'n') ||
	     (token->value.strval[0] == 'm') ||
	     (token->value.strval[0] == 'y')) &&
	    (token->value.strval[1] == 0))
	return 1;
    return 0;
}

static char get_boolean_value(TokenPtr token)
{
    if (is_boolean(token))
	return token->value.charval;
    if (is_quasi_boolean(token))
	return token->value.strval[0];
    return 0;
}

static void do_operation_or(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);

    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);

    if ((v1 == 0) || (v2 == 0))
	result = 0;

    if ((v1 == 'y') || (v2 == 'y'))
	result = 'y';
    else if ((v1 == 'm') || (v2 == 'm'))
	result = 'm';
    else
	result = 'n';

    t1->type = TYPE_TRI;
    if (t1->type == TYPE_VALUE)
	free(t1->value.strval);
    t1->value.charval = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);

    free_token(t2);
}

static void do_operation_and(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);

    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);

    if ((v1 == 0) || (v2 == 0))
	result = 0;

    if ((v1 == 'n') || (v2 == 'n'))
	result = 'n';
    else if ((v1 == 'm') || (v2 == 'm'))
	result = 'm';
    else
	result = 'y';

    t1->type = TYPE_TRI;
    if (t1->type == TYPE_VALUE)
	free(t1->value.strval);
    t1->value.charval = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);

    free_token(t2);
}

static void do_operation_not(TokenPtr *stack)
{
    TokenPtr t1;
    char result, v1;

    t1 = pop(stack);

    v1 = get_boolean_value(t1);

    if (v1 == 0)
	result = 0;

    if ((v1 == 'y') || (v1 == 'm'))
	result = 'n';
    else
	result = 'y';

    t1->type = TYPE_BOOL;
    if (t1->type == TYPE_VALUE)
	free(t1->value.strval);
    t1->value.charval = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);
}

static void do_operation_equal(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);

    /* this is just a quick and dirty hack to get
     * boolean values work. FIXME: add integer values */

    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);

    if (v1 == v2)
	result = 'y';
    else
	result = 'n';

    t1->type = TYPE_BOOL;
    if (t1->type == TYPE_VALUE)
	free(t1->value.strval);
    t1->value.charval = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);

    free_token(t2);
}

static void do_operation_not_equal(TokenPtr *stack)
{
    do_operation_equal(stack);
    do_operation_not(stack);
}

static void do_operation_lt(TokenPtr *stack)
{
}

static void do_operation_gt(TokenPtr *stack)
{
}

static void do_operation_lte(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);

    /* this is just a quick and dirty hack to get
     * boolean values work. FIXME: add integer values */

    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);

    if ((v1 == 'n') || (v2 == 'y') || ((v1 == 'm') && (v2 == 'm')))
	result = 'y';
    else
	result = 'n';

    t1->type = TYPE_BOOL;
    if (t1->type == TYPE_VALUE)
	free(t1->value.strval);
    t1->value.charval = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);

    free_token(t2);
}

static void do_operation_gte(TokenPtr *stack)
{
}

static void do_operation_depends(TokenPtr *stack)
{
    TokenPtr t1, t2;
    char result, v1, v2;

    t2 = pop(stack);
    t1 = pop(stack);

    v1 = get_boolean_value(t1);
    v2 = get_boolean_value(t2);

    if ((v1 == 0) || (v2 == 0))
	result = 'y';

    if ((v1 == 'y') && (v2 == 'n'))
	result = 'n';
    else
	result = 'y';

    t1->type = TYPE_BOOL;
    if (t1->type == TYPE_VALUE)
	free(t1->value.strval);
    t1->value.charval = result;
    t1->precedence = PRE_NULL;
    push(stack, t1);

    free_token(t2);
}

static void do_operation_derived(TokenPtr *stack, NodePtr symbol)
{
    TokenPtr t;
    char v;

    t = pop(stack);
    v = get_boolean_value(t);
    free_token(t);
    if (v == 'y')
	return;

    t = pop(stack);
    t->type = ((SymbolPtr)symbol)->type;
    t->value = ((SymbolPtr)symbol)->value;
    t->precedence = PRE_NULL;
    push(stack, t);
}

static TokenPtr duplicate_token(TokenPtr token)
{
    TokenPtr tmp;

    tmp = (TokenPtr)malloc(sizeof(Token));
    tmp->type = token->type;
    tmp->precedence = token->precedence;
    if ((tmp->type == TYPE_VALUE) || (tmp->type == TYPE_STR))
	tmp->value.strval = string_duplicate(token->value.strval);
    else
	tmp->value = token->value;
    tmp->next = token->next;

    return tmp;
}

/*
 * Expressions are represented as a sequence of tokens.  These
 * tokens have types of operators, TYPE_VALUE and TYPE_SYMBOL.
 * While evaluating, we don't alter these lists, but work on
 * copies of tokens.  However, while copying tokens of TYPE_SYMBOL
 * we also copy their type and value.
 */
TokenPtr evaluate_expression(ExpressionPtr expression, NodePtr symbol)
{
    TokenPtr token, stack, t;
    Value value;

    stack = NULL;
    value.charval = 'n';
    
    for (t = expression->first_token; t; t = t->next) {
	token = duplicate_token(t);
	switch (t->type) {
	    case TYPE_OR:
		do_operation_or(&stack);
		break;
	    case TYPE_AND:
		do_operation_and(&stack);
		break;
	    case TYPE_EQUAL:
		do_operation_equal(&stack);
		break;
	    case TYPE_NOT_EQUAL:
		do_operation_not_equal(&stack);
		break;
	    case TYPE_LT:
		do_operation_lt(&stack);
		break;
	    case TYPE_GT:
		do_operation_gt(&stack);
		break;
	    case TYPE_LTE:
		do_operation_lte(&stack);
		break;
	    case TYPE_GTE:
		do_operation_gte(&stack);
		break;
	    case TYPE_NOT:
		do_operation_not(&stack);
		break;
	    case TYPE_DEPENDS:
		do_operation_depends(&stack);
		break;
	    case TYPE_DERIVED:
		do_operation_derived(&stack, symbol);
		break;
	    case TYPE_SYMBOL:
		token->type = ((SymbolPtr)(t->value.ptr))->type;
		/* notice that we don't have to special case TYPE_STR */
		token->value = ((SymbolPtr)(t->value.ptr))->value;
		/* do not break */
	    case TYPE_VALUE:
		push(&stack, token);
		break;
	    default:
		;
	}
    }

    token = pop(&stack);
    value = token->value;

    return token;
}

/*
 * Evaluates a list of expressions pointed by EXPRESSION and retuns
 * zero if any of them evaluates to n, y otherwise.  Empty expression
 * lists are accepted.
 */
int evaluate_expressions_yn(ExpressionPtr expression, NodePtr symbol)
{
    TokenPtr token;
    ExpressionPtr e;
    char c;

    for (e = expression; e; e = e->next) {
	token = evaluate_expression(expression, symbol);
	if (((c = get_boolean_value(token)) == 0) || (c == 'n'))
	    return 0;
    }
    return 1;
}

