/*
 * KernelConf - A Linux Kernel Configuration System
 *
 * Distributed under GNU General Public Lisence
 *
 * Copyright (C) 2002 by Anuradha Ratnaweera
 */

#include <ctype.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "readsymbols.h"
#include "symbols.h"
#include "utils.h"

extern SymbolPtr symbol_tree_root;
extern SymbolPtr first_symbol;

extern MenuPtr menu_tree_root;
extern MenuPtr root_menu_item;

extern ConfFilePtr first_conf_file;

static char buffer[MAX_LINE_SIZE+1];

/*
 * Parse the next line from the config file and returns a pointer to
 * the value string in BUFFER.
 *
 * The tag type of the line is returned in TAG.
 */
static char *parse_line(FILE *conf_file, Tag *tag)
{
    int i;
    char *bufptr, c;

    /* let's read the next line from SYMBOLFILE */
    i = 0;
    while (((c = fgetc(conf_file)) != EOF) && (c != '\n') && (i < MAX_LINE_SIZE))
	buffer[i++] = c;
    buffer[i] = 0;

    if ((c == EOF) && (i == 0)) {
	*tag = TAG_EOF;
	return NULL;
    }

    if (i == MAX_LINE_SIZE) {
	fprintf(stderr,
		"Warning: Line with length %d read, may have been truncated\n",
		MAX_LINE_SIZE);
    }

    /* remove comments */
    if ((bufptr = strchr(buffer, '#')))
	*bufptr = 0;

    /* if first line is space, we may have a description
     * line or a blank line */
    if (isspace(buffer[0])) {
	bufptr = string_trim(buffer);
	if (*bufptr) {
	    *tag = TAG_DESC_LINE;
	    return bufptr;
	}
	else {
	    *tag = TAG_BLANK;
	    return NULL;
	}
    }
    if (buffer[0] == 0) {
	*tag = TAG_BLANK;
	return NULL;
    }
    /* now we handle individual tags */
    if (strncmp(buffer, "include", 7) == 0) {
	*tag = TAG_INCLUDE;
	return string_trim(buffer + 7);
    }
    if (strncmp(buffer, "symbol:", 7) == 0) {
	*tag = TAG_SYMBOL;
	return string_trim(buffer + 7);
    }
    if (strncmp(buffer, "type:", 5) == 0) {
	*tag = TAG_TYPE;
	return string_trim(buffer + 5);
    }
    if (strncmp(buffer, "value:", 6) == 0) {
	*tag = TAG_VALUE;
	return string_trim(buffer + 6);
    }
    if (strncmp(buffer, "default:", 8) == 0) {
	*tag = TAG_DEFAULT;
	return string_trim(buffer + 8);
    }
    if (strncmp(buffer, "description:", 12) == 0) {
	*tag = TAG_DESC;
	return string_trim(buffer + 12);
    }
    if (strncmp(buffer, "depends:", 8) == 0) {
	*tag = TAG_DESC;
	return string_trim(buffer + 8);
    }
    if (strncmp(buffer, "derived:", 8) == 0) {
	*tag = TAG_DESC;
	return string_trim(buffer + 8);
    }
    if (strncmp(buffer, "menu-item:", 10) == 0) {
	*tag = TAG_MENU_ITEM;
	return string_trim(buffer + 10);
    }
    if (strncmp(buffer, "menu-caption:", 13) == 0) {
	*tag = TAG_MENU_CAPTION;
	return string_trim(buffer + 13);
    }
    if (strncmp(buffer, "menu-parent:", 12) == 0) {
	*tag = TAG_MENU_PARENT;
	return string_trim(buffer + 12);
    }
    if (strncmp(buffer, "menu-depends:", 13) == 0) {
	*tag = TAG_MENU_DEPENDS;
	return string_trim(buffer + 13);
    }
    /* unknown tag, return NULL */
    *tag = TAG_NULL;
    return NULL;
}

/*
 * Adds MENU_ITEM to the end of PARENT's children list.
 * The next element of MENU_ITEM is assumed to be NULL.
 */
static void add_menu_to_parent(MenuPtr menu_item, MenuPtr parent)
{
    NodePtr ptr;
    if (parent->child) {
	ptr = (NodePtr)parent->child;
	while (ptr->next)
	    ptr = ptr->next;
	ptr->next = (NodePtr)menu_item;
    }
    else {
	parent->child = menu_item;
    }
}

static ConfFilePtr add_conf_file(char *filename)
{
    ConfFilePtr ptr;

    ptr = first_conf_file;
    while (ptr && ptr->next)
	ptr = ptr->next;

    if (ptr) {
	ptr->next = (ConfFilePtr)malloc(sizeof(ConfFile));
	ptr = ptr->next;
    }
    else {
	ptr = (ConfFilePtr)malloc(sizeof(ConfFile));
	first_conf_file = ptr;
    }

    ptr->name = string_duplicate(filename);
    ptr->next = NULL;

    return ptr;
}

/*
 * Reads configuration symbols, menus and dependencies from .conf
 * files in the kernel tree
 */
int read_symbols(char *filename)
{
    Tag tag;
    FILE *conf_file;
    char *buf;
    int line_no;
    char last_line_blank, in_description;
    SymbolPtr symbol = NULL;
    static SymbolPtr last_symbol = NULL;
    MenuPtr menu_item = NULL;
    ConfFilePtr this_file;
    long location;

#ifdef DEBUG
    printf("Processing file: %s\n", filename);
#endif

    if ((conf_file = fopen(filename, "rt")) == NULL) {
	fprintf(stderr, "Error: can't open %s\n", filename);
	exit(1);
    }

    this_file = add_conf_file(filename);

    line_no = 0;
    last_line_blank = 1;

    if (root_menu_item == NULL) {
	root_menu_item = (MenuPtr)search_tree((NodePtr*)(&menu_tree_root),
		"root", NODE_INSERT, sizeof(Menu), NULL);
	root_menu_item->parent = NULL;
	root_menu_item->child = NULL;
	((NodePtr)root_menu_item)->next = NULL;
	root_menu_item->caption = string_duplicate("Main menu");;
	root_menu_item->depends = NULL;
	root_menu_item->symbol = NULL;
	root_menu_item->conf_file = NULL;
	root_menu_item->desc_start = -1;
	root_menu_item->desc_length = 0;
    }
	
    do {
	buf = parse_line(conf_file, &tag);
	line_no++;

	switch (tag) {
	    case TAG_NULL:
		fprintf(stderr, "Warning: %s:%d: unknown tag\n",
			filename, line_no);
		break;
	    case TAG_INCLUDE:
		read_symbols(buf);
		symbol = NULL;
		menu_item = NULL;
		break;
	    case TAG_SYMBOL:
		if (last_line_blank) {
		    menu_item = NULL;
		    symbol = (SymbolPtr)search_tree((NodePtr*)(&symbol_tree_root),
			    buf, NODE_INSERT, sizeof(Symbol), (NodePtr)last_symbol);
		    last_symbol = symbol;
		    if (first_symbol == NULL)
			first_symbol = symbol;
		    symbol->type = TYPE_BOOL;
		    symbol->value.charval = 'n';
		    symbol->depends = NULL;
		    symbol->derived = NULL;
		    symbol->menu_item = NULL;
		    symbol->conf_file = this_file;
		    symbol->desc_start = -1;
		    symbol->desc_length = 0;
		}
		else {
		    fprintf(stderr,
			    "%s:%d: Symbol defined in the middle of the section\n",
			    filename, line_no);
		    exit(1);
		}
		break;
	    case TAG_TYPE:
		if (symbol == NULL) {
		    fprintf(stderr,
			    "%s:%d: Type definition without symbol\n",
			    filename, line_no);
		    exit(1);
		}
		if (strncmp(buf, "bool", 4) == 0) {
		    symbol->type = TYPE_BOOL;
		}
		else if (strncmp(buf, "trist", 5) == 0) {
		    symbol->type = TYPE_TRI;
		}
		else if (strncmp(buf, "int", 3) == 0) {
		    symbol->type = TYPE_INT;
		    symbol->value.intval = 0;
		}
		else if (strncmp(buf, "hex", 3) == 0) {
		    symbol->type = TYPE_HEX;
		    symbol->value.intval = 0;
		}
		else if (strncmp(buf, "char", 4) == 0) {
		    symbol->type = TYPE_CHAR;
		}
		else if (strncmp(buf, "str", 3) == 0) {
		    symbol->type = TYPE_STR;
		    symbol->value.strval = NULL;
		}
		else if (strncmp(buf, "option", 6) == 0) {
		    symbol->type = TYPE_OPTION;
		}
		break;
	    case TAG_VALUE:
	    case TAG_DEFAULT:
		if (symbol == NULL) {
		    fprintf(stderr, "%s:%d: Value defined without symbol\n",
			    filename, line_no);
		    exit(1);
		}
		switch (symbol->type) {
		    case TYPE_BOOL:
		    case TYPE_OPTION:
			if ((buf[0] == 'Y') || (buf[0] == 'y'))
			    symbol->value.charval = 'y';
			else
			    symbol->value.charval = 'n';
			break;
		    case TYPE_TRI:
			if ((buf[0] == 'Y') || (buf[0] == 'y'))
			    symbol->value.charval = 'y';
			else if ((buf[0] == 'M') || (buf[0] == 'm'))
			    symbol->value.charval = 'm';
			else
			    symbol->value.charval = 'n';
			break;
		    case TYPE_INT:
		    case TYPE_HEX:
			symbol->value.intval = atoi(buf);
			break;
		    case TYPE_CHAR:
			symbol->value.intval = buf[0];
			break;
		    case TYPE_STR:
			symbol->value.strval = string_duplicate(buf);
			break;
		    default:
			;
		}
		break;
	    case TAG_DEPENDS:
		if (symbol == NULL) {
		    fprintf(stderr,
			    "%s:%d: Dependency without symbol\n",
			    filename, line_no);
		    exit(1);
		}
		symbol->depends = parse_expressions(buf, (NodePtr)symbol, 1);
		break;
	    case TAG_DERIVED:
		if (symbol == NULL) {
		    fprintf(stderr,
			    "%s:%d: Derived value without symbol\n",
			    filename, line_no);
		    exit(1);
		}
		symbol->derived = parse_expressions(buf, (NodePtr)symbol, 0);
		break;
	    case TAG_MENU_ITEM:
		if (last_line_blank)
		    symbol = NULL;
		menu_item = (MenuPtr)search_tree((NodePtr*)(&menu_tree_root),
			buf, NODE_INSERT, sizeof(Menu), NULL);
		menu_item->parent = NULL;
		menu_item->child = NULL;
		((NodePtr)menu_item)->next = NULL;
		menu_item->caption = NULL;
		menu_item->depends = NULL;
		menu_item->symbol = symbol;
		if (symbol)
		    symbol->menu_item = menu_item;
		menu_item->conf_file = this_file;
		menu_item->desc_start = -1;
		menu_item->desc_length = 0;
		break;
	    case TAG_MENU_CAPTION:
		if (menu_item == NULL) {
		    fprintf(stderr,
			    "%s:%d: Menu caption defined without menu item\n",
			    filename, line_no);
		    exit(1);
		}
		menu_item->caption = string_duplicate(buf);
		break;
	    case TAG_MENU_PARENT:
		if (menu_item == NULL) {
		    fprintf(stderr,
			    "%s:%d: Menu parent defined without menu item\n",
			    filename, line_no);
		    exit(1);
		}
		menu_item->parent = (MenuPtr)search_tree((NodePtr*)(&menu_tree_root),
			buf, NODE_MARK_INSERT, sizeof(Menu), NULL);
		add_menu_to_parent(menu_item, menu_item->parent);
		break;
	    case TAG_MENU_DEPENDS:
		if (menu_item == NULL) {
		    fprintf(stderr,
			    "%s:%d: Menu dependency defined without menu item\n",
			    filename, line_no);
		    exit(1);
		}
		menu_item->depends = parse_expressions(buf,
			(NodePtr)(menu_item->symbol), 0);
		break;
	    case TAG_BLANK:
	    case TAG_EOF:
		if (menu_item && (menu_item->parent == NULL)) {
		    menu_item->parent = root_menu_item;
		    add_menu_to_parent(menu_item, root_menu_item);
		}
		break;
	    case TAG_DESC:
		in_description = 1;
		location = ftell(conf_file);
		if (symbol)
		    symbol->desc_start = location;
		if (menu_item)
		    menu_item->desc_start = location;
		break;
	    case TAG_DESC_LINE:
		if (in_description) {
		    location = ftell(conf_file);
		    if (symbol)
			symbol->desc_length = location - symbol->desc_start;
		    if (menu_item)
			menu_item->desc_length = location - menu_item->desc_start;
		}
		break;
	    default:
		;
	}

	if (tag == TAG_BLANK)
	    last_line_blank = 1;
	else
	    last_line_blank = 0;

	if ((tag != TAG_BLANK) && (tag != TAG_DESC) && (tag != TAG_DESC_LINE))
	    in_description = 0;

    } while (tag != TAG_EOF);

    fclose(conf_file);

    return 0;
}

